import React, {useEffect, useRef, useState} from "react";
import {ColumnKey} from "./Column";
import {normalizeRowDefinition, RowData, RowDefinition} from "./Row";
import {CellDefinition} from "./Cell";
import {SmartableData} from "./Smartable";
import {Modify, Promisable} from "@kernelui/core";

/**
 * Current Smartable data state.
 */
export interface CurrentTableData<CK extends ColumnKey>
{
	/**
	 * Current rows state, undefined if they are still loading.
	 */
	rows?: CurrentRowData<CK>[];
}

/**
 * Smartable current row cells data state.
 */
export type CurrentRowData<CK extends ColumnKey, T = any> = Modify<RowData<CK>, {
	cells: Record<CK, CellDefinition<T>>;
}>;


/**
 * Get current async table data manager.
 */
export function useAsyncManager<CK extends ColumnKey>(data: SmartableData<CK>): AsyncManager<CK>
{
	// Get the main instance of async manager.
	const asyncManager = useRef<AsyncManager<CK>>();

	// Get the main async manager state.
	const [currentDataState, setCurrentDataState] = useState<CurrentTableData<CK>>({
		rows: undefined,
	});

	if (!asyncManager.current)
	{
		// Initialize a new async manager if there is none.
		asyncManager.current = new AsyncManager<CK>();
	}

	// Update current data state and its dispatcher.
	asyncManager.current.currentDataState = currentDataState;
	asyncManager.current.setCurrentDataState = setCurrentDataState;

	// When defined table data change, process async state again.
	useEffect(() => {
		// Process new data.
		asyncManager.current.handle(data);
	}, [asyncManager, data]);

	// Return the main instance of async manager.
	return asyncManager.current;
}


/**
 * Smartable async data manager.
 */
class AsyncManager<CK extends ColumnKey>
{
	currentDataState: CurrentTableData<CK>;
	setCurrentDataState: React.Dispatch<React.SetStateAction<CurrentTableData<CK>>>;

	/**
	 * Main promised data.
	 * @protected
	 */
	protected promisedData: Promised<SmartableData<CK>>;

	/**
	 * Promised rows data.
	 * @protected
	 */
	protected promisedRows: Promised<RowDefinition<CK>>[] = [];

	/**
	 * Promised rows cells data.
	 * @protected
	 */
	protected promisedRowsCells: Partial<Record<CK, Promised<CellDefinition>>>[] = [];


	/**
	 * Tells if rows are loaded or not.
	 * @protected
	 */
	protected rowsLoaded: boolean = false;

	/**
	 * Rows data.
	 * @protected
	 */
	protected rowsData: RowData<CK>[] = [];

	/**
	 * Rows cells definitions.
	 * @protected
	 */
	protected cellsDefinitions: Partial<Record<CK, CellDefinition<CK>>>[] = [];

	constructor()
	{
		// Initialize promised data object.
		this.promisedData = new Promised(this.handleNewRowsDefinitions.bind(this));
	}

	/**
	 * Handle new smartable data.
	 * @param data Smartable data.
	 */
	handle(data: SmartableData<CK>): void
	{
		this.rowsLoaded = false;
		this.promisedData.refresh(data);
	}

	/**
	 * Handle new rows definitions.
	 * @param rowsDefinitions Rows definitions.
	 * @protected
	 */
	protected handleNewRowsDefinitions(rowsDefinitions: Promisable<RowDefinition<CK>>[]): void
	{
		// Ignore undefined value.
		if (rowsDefinitions == undefined) return;

		// Initialize rows data and cells definitions.
		this.rowsData = [];
		this.cellsDefinitions = [];
		this.rowsLoaded = true;

		for (const [rowId, rowDefinition] of rowsDefinitions.entries())
		{ // Get row data of each row.
			if (!this.promisedRows[rowId])
				this.promisedRows[rowId] = new Promised(this.handleNewRow.bind(this, rowId), rowDefinition);
			else
				this.promisedRows[rowId].refresh(rowDefinition);
		}

		// Try to update the current data state.
		this.tryToUpdate();
	}

	/**
	 * Handle a new row.
	 * @param rowId Row ID.
	 * @param row Row definition.
	 * @protected
	 */
	protected handleNewRow(rowId: number, row: RowDefinition<CK>): void
	{
		// Ignore undefined value.
		if (row == undefined) return;

		// Normalize row data.
		const rowData = normalizeRowDefinition(row);

		// Save row data.
		this.rowsData[rowId] = rowData;
		// Initialize cells definition.
		this.cellsDefinitions[rowId] = {};

		for (const [columnKey, cellDefinition] of Object.entries(rowData.cells) as [CK, CellDefinition][])
		{ // Get cell definition of each row cell.
			if (!this.promisedRowsCells[rowId]) this.promisedRowsCells[rowId] = {};
			if (!this.promisedRowsCells[rowId][columnKey])
				this.promisedRowsCells[rowId][columnKey] = new Promised(this.handleNewCell.bind(this, rowId, columnKey), cellDefinition);
			else
				this.promisedRowsCells[rowId][columnKey].refresh(cellDefinition);
		}

		// Try to update the current data state.
		this.tryToUpdate();
	}

	/**
	 * Handle a new cell.
	 * @param rowId Cell row ID.
	 * @param columnKey Cell column key.
	 * @param cellDefinition Cell definition.
	 * @protected
	 */
	protected handleNewCell(rowId: number, columnKey: CK, cellDefinition: CellDefinition): void
	{
		// Ignore undefined value.
		if (cellDefinition == undefined) return;

		// Save cell definition.
		this.cellsDefinitions[rowId][columnKey] = cellDefinition;

		// Try to update the current data state.
		this.tryToUpdate();
	}

	/**
	 * True if there is a pending update.
	 * @protected
	 */
	protected pendingUpdate: boolean = false;

	/**
	 * Try to update the current data state, if there are no more changes in the next 25ms.
	 * The next update will be in max 25ms.
	 * @protected
	 */
	protected tryToUpdate(): void
	{
		if (!this.pendingUpdate)
			// Try to update in the next 25ms.
			setTimeout(() => {
				this.pendingUpdate = false;
				this.update();
			}, 25);
	}

	/**
	 * Update the current data state with the loaded rows and cells data.
	 * @protected
	 */
	protected update(): void
	{
		if (!(
			// Checking that there is at least one changed value.
			this.rowsData.length > 0 || this.cellsDefinitions.some((rowCells) => (
				Object.keys(rowCells).length > 0
			))
			|| ((this.rowsLoaded && !this.currentDataState.rows) || (!this.rowsLoaded && this.currentDataState.rows))
		)) // Nothing has changed.
			return;

		// Initialize new data.
		const newData = {
			rows: !this.rowsLoaded ? undefined : [
				...(this.currentDataState?.rows ?? [])
			],
		};

		for (const [rowId, newRow] of this.rowsData?.entries())
		{ // Update value of each new row.
			newData.rows[rowId] = {
				element: newRow.element,
				cellElement: newRow.cellElement,
				cells: newData.rows[rowId]?.cells,
			};
		}

		for (const [rowId, rowCells] of this.cellsDefinitions?.entries())
		{ // Update cells of each changed row.
			newData.rows[rowId] = {
				...newData.rows[rowId],
				cells: {...(newData.rows[rowId]?.cells ?? {}), ...rowCells} as Record<CK, CellDefinition>,
			}
		}

		// Update the current data state.
		this.currentDataState = newData;
		this.setCurrentDataState(
			newData
		);
	}
}

/**
 * Promised data class.
 */
export class Promised<T>
{
	/**
	 * The main data promise.
	 */
	promise?: Promise<T>;

	/**
	 * Data retrieved from promise or given in parameter.
	 */
	data?: T;

	/**
	 * Called when data is changed.
	 */
	onData?: (data: T) => void;

	constructor(onChanged?: (data: T) => void, data?: Promisable<T>)
	{
		this.onData = onChanged;
		if (data)
			this.refresh(data);
	}

	/**
	 * Refresh the promised data.
	 * @param data Promised data.
	 */
	refresh(data: Promisable<T>): this
	{
		if (data instanceof Promise)
		{ // We have a promise of data.
			if (data != this.promise)
			{ // The promise is different from the saved one.
				// Save the new promise and set data to undefined.
				this.promise = data;
				this.data = undefined;
				this.onData?.(undefined);

				// Wait for promise to resolve to get actual data.
				this.promise.then((data) => {
					// Data is retrieved, saving it.
					this.data = data;
					this.onData?.(data);
				});
			}
		}
		else if (data != this.data)
		{ // We already have data, and it is different from the current state.
			this.data = data;
			this.onData?.(data);
		}

		return this;
	}

	/**
	 * Return true if some data (or its promise) have been provided.
	 */
	isInitialized(): boolean
	{
		return !!this.data || !!this.promise;
	}

	/**
	 * Return true if we are waiting for a promise result.
	 */
	isLoading(): boolean
	{
		return this.data === undefined && !!this.promise;
	}
}
