import React, {useCallback} from "react";
import {ColumnFilter, useFilterState} from "../Columns/ColumnFilter";
import {normalizeString} from "@kernelui/core";

/**
 * Filter value regex.
 */
export const filterRegex = /^([=!])?([^=!].+)$/;

/**
 * String column filter.
 */
export const StringFilter: ColumnFilter<string, StringFilterState> = {
	filter: (data: string, filterState: StringFilterState) => {
		// Read filter value.
		const filterValue = filterRegex.exec(filterState.value);
		// No valid filter value, allow everything.
		if (!filterValue?.length) return true;
		// Get current filter result.
		const filterResult = normalizeString(data).includes(normalizeString(filterValue[2]));
		// Invert filter result based on filter modifier.
		return filterValue[1] == "!" ? !filterResult : filterResult;
	},

	element: <StringFilterComponent />
};

/**
 * String filter state.
 */
export interface StringFilterState
{
	/**
	 * Filter value.
	 */
	value: string;
}

/**
 * String filter component.
 */
export function StringFilterComponent()
{
	// Initialize string filter state.
	const [stringFilterState, setStringFilterState] =
		useFilterState<StringFilterState>({ value: "" });

	// Handle filter input change.
	const handleChange = useCallback((event: React.ChangeEvent<HTMLInputElement>) => {
		// Save the current filter value.
		setStringFilterState({
			value: event.currentTarget.value,
		})
	}, [setStringFilterState]);

	return (
		<input type={"text"} pattern={filterRegex.source} size={1}
		       value={stringFilterState.value} onChange={handleChange} />
	);
}
