import React, {useMemo} from "react";
import {AutoColumnContextProvider, Column, ColumnKey, Columns} from "./Column";
import {SmartableProperties, useTable} from "./Smartable";
import {RowInstance, RowLoader} from "./Row";
import {useAsyncManager} from "./AsyncManager";
import {ColumnHeading} from "./Columns/ColumnHeading";
import {sortRows} from "./Sort";
import {AutoPaginate} from "@kernelui/core";

/**
 * Smartable instance component properties.
 */
export interface InstanceProperties<CK extends ColumnKey> extends SmartableProperties<CK>
{
	/**
	 * Table columns.
	 */
	columns: Columns<CK>;
}

/**
 * Main component for a Smartable table.
 */
export function Instance<CK extends ColumnKey>(props: InstanceProperties<CK>)
{
	if (props.paginate)
	{ // If pagination is enabled.
		return <PaginatedInstance {...props} />
	}
	else
	{ // No pagination, simple body render.
		return (
			<Table {...props}>
				<TableBody />
			</Table>
		);
	}
}

/**
 * Paginated Smartable instance component.
 */
export function PaginatedInstance<CK extends ColumnKey>(props: InstanceProperties<CK>)
{
	// Get data from table.
	const {data} = useTable<CK>();

	// Get current data state from the async table value.
	const {currentDataState} = useAsyncManager<CK>(data);

	// Compute page count for the current rows.
	const pageCount = Math.ceil((currentDataState?.rows?.length ?? 0) / props.paginate.pageSize);

	return (
		<AutoPaginate count={pageCount}>
			{(page) => (
				// Render table with table body of the current page.
				<Table {...props}>
					<TableBody pagination={{page: page, pageSize: props.paginate.pageSize}} />
				</Table>
			)}
		</AutoPaginate>
	);
}

/**
 * Base component for a Smartable table.
 */
export function Table<CK extends ColumnKey>({columns, children}: React.PropsWithChildren<InstanceProperties<CK>>)
{
	return (
		<table className={"smartable"}>
			<thead>
			<ColumnsHeadings columns={columns} />
			</thead>
			<tbody>
			{children}
			</tbody>
		</table>
	);
}

/**
 * Columns headings of a Smartable table.
 */
export function ColumnsHeadings<CK extends ColumnKey>({columns}: {columns: Columns<CK>})
{
	return (
		<>
			<tr className={"headings"}>
				{ // Showing title of each column.
					Object.keys(columns).map((key) => (
						<AutoColumnContextProvider key={key} columnKey={key}>
							<ColumnHeading/>
						</AutoColumnContextProvider>
					))
				}
			</tr>
			<tr className={"filters"}>
				{ // Add columns filters, if there are some.
					(Object.entries(columns) as [CK, Column][]).map(([columnKey, column]) => (
						column.filter && (
							<AutoColumnContextProvider key={columnKey as string} columnKey={columnKey}>
								<td>{column.filter.element}</td>
							</AutoColumnContextProvider>
						)
					))
				}
			</tr>
		</>
	);
}

/**
 * Smartable table body.
 */
export function TableBody<CK extends ColumnKey>({pagination}: {
	/**
	 * Current pagination state.
	 */
	pagination?: { page: number; pageSize: number; };
})
{
	// Get data from table.
	const {data, columns, columnsSortState, columnsFilterStates} = useTable<CK>();

	// Get current data state from the async table value.
	const {currentDataState} = useAsyncManager<CK>(data);

	// Memorize filtered rows.
	const filteredRows = useMemo(() => (
		currentDataState.rows?.filter((row) => (
			// Checking each row to keep only those which match the filters.
			(Object.entries(columnsFilterStates) as [CK, any][]).every(([columnKey, filterState]) => (
				// For each filter, keep the row if data match the current filter.
				columns[columnKey].filter.filter(row.cells[columnKey].data, filterState)
			))
		))
	), [currentDataState.rows, columnsFilterStates]);

	// Memorize sorted rows.
	let sortedRows = useMemo(() => (
		// Sort rows with the current columns sort state.
		sortRows<CK>(filteredRows, columns, columnsSortState)
	), [filteredRows, columns, columnsSortState]);

	if (pagination)
	{ // If pagination is enabled, showing only content of the current page.
		const startIndex = (pagination.page - 1) * pagination.pageSize;
		sortedRows = sortedRows?.slice(startIndex, startIndex + pagination.pageSize);
	}

	return (
		sortedRows ? (
			sortedRows.map((rowData, index) => (
				// Rendering each row from its definition.
				<RowInstance row={rowData} />
			))
		) : (
			<RowLoader />
		)
	);
}
